# Copyright (C) 2006, Augusto Ciuffoletti <augusto@di.unipi.it>
# $Revision: 1.32 $
# The following comes from CPAN (thanks Brad Fitzpatrick)
use strict;
use Crypt::OpenSSL::DSA;
use Time::HiRes qw( usleep ualarm gettimeofday tv_interval );
use peerdb;

=head1 NAME
stoken - Secure token passing protocol library

=head1 SYNOPSIS

B<use lib "stoken">;

=head1 Description

The stoken library supports the secure exchange of a token with a
trusted peer. The protocol ensures that the token is never
duplicated. The token loss event is characterized by a low probability
to occur.

The token passing operation is carried out in three steps, between a source and a destination interfaces:

=over 4

=item * B<move>: the token is sent from the source to the interface;

=item * B<acknowledge>: the destination confirms the source the
receipt of the token;

=item * B<commit>: the source confirms the destination the release of
the token.

=item * B<earlystop>: UNDOCUMENTED.

=back

The sender holds the token until it sends the commit message, and the
destination holds it as soon as it receives that same message.

The token does not carry user data, but a distinguishing
identifier. Upon receipt, some statistics about the token are made
available to the destination, besides the token identifier:

=over 4

=item * B<generation timestamp>:(todo) the time when the token was first generated;

=item * B<generation interface>:(todo) the interface that first generated the token;

=item * B<token latency>: an upper bound of the time taken by the token passing operation;

=item * B<source interface>: the interface from where the token comes from.

=back

Security is obtained using a DSA signature of token contents. The
signature is generated using source private key. The library provides,
transparently embedded in token passing subroutines, the signature
generation as well as the public key passing functionalities.

In order to ensure a low footprint, the public key is passed only in
case it is not yet known to the destination.

=head2 Level 0: Key management

=head3 add_certificate($msg)

Appends a public key or certificate to an already formed packet.

Parameters:

=over 4

=item * B<msg>: the message, a null terminated string;

=back 4

=head3 check_token($peer,$token,$signature)

Check validity of a token using packet signature and peer identity.

Parameters:

=over 4

=item * B<peer>: the identity of the peer, null terminated string with
IP address; 

=item * B<token>: the token to be validated, a null terminated string;

=item * B<signature>: the signature attached to the token, a null terminated string;

=back 4

Returns:

=over 4

=item * 0 upon successful validation

=item * 1 if one of the following errors occurs:

=over 4 

=item - if peer's keyfile does not exixts, 
    
=item - if peer's keyfile is malformed (in that case keyfile is removed), 
    
=item - if authentication fails.

=back

=back 4

=head2 Level 1: Packet authentication

=head3 build_packet($format, $flags,$timing,$token_id,$token_passing_id)

Takes the values of the fields contained in the packet, generates a
signature, and returns a signed packet, obtained packing together the
components as passed by parameters using the packing format passed as
first parameter. The signature is computed from fields that remain
unaltered during token exchange (security hole...).

A key pair is generated when keys are not existent (at startup) or when they
are malformed.

Parameters:

=over 4

=item * B<format>: the format of the packet (signature excluded), 
as a "pack" string.

=item * B<field values>: The values of the fields to be packed.

=back 4

Returns:

The packet ready to be sent.

=head3 parse_packet($format,$msg,$hispaddr)

Takes as parameter a packet, and checks it using the public key
available for the source interface.  In case the packet contains a
public key, the key is stored as corresponding to the packet source
interface.  

Parameters:

=over 4

=item * B<format>: the format of the packet (signature excluded), 
as a "pack" string.

=item * B<msg>: the incoming message.

=item * B<hispaddr>: the port address of the peer 

=back 4

Returns:

Returns an array of the unpacked content of the packet upon success, or undef
if token validtion fails.

=head2 Level 2: Token Passing primitives

=head3 token_wait($port, $wait_timeout, $commit_timeout, $retry_timeout)

Waits for a token on a specific port, and returns the token
identifier, the wait interval length, and the token passing operation
duration when a successful token passing occurs. If no token is received
within the B<wait_timeout>, it returns undefined.

When a token is received, an ack is sent. After sending the ack, it
keeps listening for a commit during an interval of duration
B<commit_timeout>. During that time, every B<retry_timeout> a new ack
is sent. When the B<commit_timeout> expires, it restarts waiting for
another token, without returning.

Parameters:

=over 4

=item * B<port>: where to wait for the incoming packet.

=item * B<wait_timeout>: how long to wait for a new packet.

=item * B<commit_timeout>: how long to wait for the commit before failure.

=item * B<retry_timeout>: how long to wait for a commit before retrying.

=back 4

Returns:

A list containing 

=item * B<token_id>: the id of the received token.

=item * B<waiting_time>: the duration of the interval from the last token receive.

=item * B<token_latency>: an upper bound of the time between the send of the first packet (move) and the receipt of the last packet (commit).

=back 4

=head3 token_pass( $token_id, $peer, $port, $ack_timeout, $retry_timeout)

Passes a token to a peer on the (well known) port, and returns 1
(true) if the token is passed successfully. Returns undef if no ack is
received within the B<ack_timeout>, of if the commit sending failed.

After sending the move packet, it keeps listening for an ack during an
interval of duration B<ack_timeout>. During that time, every
B<retry_timeout> a new move packet is sent. When the B<ack_timeout>
expires, the subroutine returns with an undef value. If an ack is
received during that interval, it is evaluated for sanity, and a comit
message is returned to the sender.

Parameters:

=over 4

=item * B<token_id>: the token to pass.

=item * B<peer>: the target peer.

=item * B<port>: the well known port where to send the token.

=item * B<ack_timeout>: how long to wait for an ack before failure.

=item * B<retry_timeout>: how long to wait for an ack retrying.

=back 4

Returns:

True upon success, undef otherwise.

=cut

my $VERSION='$Revision: 1.32 $';
$VERSION =~ s/\$//g;
print "using stoken.pm: $VERSION\n";

my $certdir="/tmp";
my $private_key_filename=$certdir."/id_dsa.priv";
my $public_key_filename=$certdir."/id_dsa.pub";
my $stacksize=10;
my $debug=1;
my $socket;
my $peer_database;
$|=1;

my @digits=('0','1','2','3','4','5','6','7','8','9','a','b','c','d','e','f');

BEGIN {
	$peer_database=peerdb->new;
}

sub logging {
    printf "%14.3f: ",Time::HiRes::time();
    printf @_;
}

sub soft_stop {
	logging "! Closing receive socket\n";
	( defined $socket ) && close $socket;
}

#sub clear_pkdb {
#    opendir DB,$certdir;
#    while ( my $fn=readdir(DB) ) {
#	( $fn =~ /\d+\.\d+\.\d+\.\d+.pub/ ) || next;
#	unlink "$certdir/$fn";
#    }
#    unlink "$certdir/id_dsa.pub";
#    unlink "$certdir/id_dsa.priv";
#    closedir DB;
#}

sub randid {
    my $result="";
    for (my $i=0;$i<8;$i++) {
	$result.=$digits[int(rand(16))];
    }
    return $result;
}

sub randport {
    return 12000+int(rand(100));
}

# THIS IS FOR DEBUGGING...
sub dump_packet {
    my ( $header,$msg ) = @_;
    printf "=== $header (length %d octets) === \n",length($msg);;
    for (my $i=0; $i<length($msg); $i++) {
	logging "%02x ",ord(substr($msg,$i,1));
    }
    print "\n===\n";
}

###################
#
# Level 0: Keys management
#
###################

sub add_certificate {
# Appends a public key (or a certificate) to an already formed packet
    my ( $msg ) = @_;
    open (PUBKEY,$public_key_filename);
    while ( <PUBKEY> ) { $msg .= $_ };
    close PUBKEY;
    return $msg;
}

sub generate_keys {
    my $dsa = Crypt::OpenSSL::DSA->generate_parameters( 512 );
    $dsa->generate_key;
    $dsa->write_priv_key( $private_key_filename );
    $dsa->write_pub_key( $public_key_filename );
}

#sub record_peer_key {
#    my ( $peer, $public_key )=@_;
#    open(PUBKEY,">$certdir/$peer.pub");
#    print PUBKEY $public_key;
#    close PUBKEY;
#}

sub check_token {
    my ($peer,$token,$signature)=@_;
	my $dsa_pub=$peer_database->get_certificate($peer);
#    my $keyfile=$certdir."/".$peer.".pub";
#    if ( ! -f $keyfile ) { 
#	logging "!  Authentication failure: keyfile $keyfile does not exist\n";
#	return 0;
#    }
#    my $dsa_pub=Crypt::OpenSSL::DSA->read_pub_key($keyfile);
    if ( ! defined $dsa_pub ) {
	logging "!  Authentication failure with $peer\n";
	return 0;
    }
# Check token authentication
    if ( ! $dsa_pub->verify($token, $signature) ) {
	logging "!  Authentication failed with $peer";
	return 0;
    }
    return 1;
}

###################
#
# Level 1: Packet management
#
###################

sub build_packet { 
    my ( $format, $flags, $timing, $token_id, $token_passing_id ) = @_;

# Generate keys and files in not present...
    if ( ! -f $private_key_filename || ! -f $public_key_filename ) {
        logging "!  Local keys do not exists: regenerating\n";
	generate_keys();
    }
# ...or malformed
    my $dsa_priv = Crypt::OpenSSL::DSA->read_priv_key( $private_key_filename );
    if ( ! defined $dsa_priv ) {
	logging "!  Malformed private key: regenerating\n";
	generate_keys();
    }
# Compute signature
    my $sig = $dsa_priv->sign($token_id.$token_passing_id);
# Pack packet
    my $msg=pack($format."n", $flags, $timing,$token_id,$token_passing_id, length($sig));
    $msg.=$sig;
    return ( $msg );
}

sub parse_packet {
    my ( $format, $msg, $hispaddr ) = @_;

    logging "   * Parsing packet\n";
# Extract peer address
    my ( $peer_port, $hisiaddr) = sockaddr_in($hispaddr);
    my $peer=inet_ntoa($hisiaddr);
# Extract token and operation identifiers
    my ($flags,$timing,$token_id,$token_passing_id,$siglength)=
	unpack($format."n",$msg);
# Extract signature
    my $signature=substr($msg,length(pack($format."n")));
# Check message type
    if ( length($signature) != $siglength ) {
	logging "            (contains pub key)\n";
	my $public_key=unpack("a*",substr($signature,$siglength));
	$peer_database->add($peer,$public_key);
    }
    check_token($peer,$token_id.$token_passing_id,$signature) || return undef;

    return ( $flags, $timing, $token_id,$token_passing_id );
}

#######################
#
# LEVEL 2: token passing operations
#
#######################

    

sub mkSocket {
    my ($port,$hisiaddr) = @_;
    socket(SOCKET, PF_INET, SOCK_DGRAM, getprotobyname('udp')) 
	|| die "socket: $!";
    binmode(SOCKET,":raw");
    bind(SOCKET, sockaddr_in($port, $hisiaddr))
	|| die "bind: $!";
    return *SOCKET;
}

##
# DEBUG ONLY
##
sub inject_problem {
    my $r=rand(100);
    if ( $r<5 ) { 
	die "*\n* injected loss\n*\n";
    };
    my $t=80;
    if ( $r>$t ) {
	my $lapse=int($t/($r-$t+1));
	logging "*\n* injected $lapse sec delay\n*\n"; 
	sleep $lapse; 
    }
}    

sub sendMove {
    my ($socket,$token_id,$token_passing_id,$lost,$peerpaddr)=@_;
    my $flag=0;
    eval {
	logging " > Sending MOVE packet (%s, %s)\n"
	    ,$token_id, $token_passing_id;
	( $debug == -1 ) && inject_problem();
	my $msg=build_packet("Cnh8h8",$flag,0,$token_id,$token_passing_id);
	( $lost > 0 ) && ( $msg = add_certificate($msg) );	    
	( $debug > 2 ) && dump_packet("MOVE",$msg);
# Send MOVE packet
	defined( send($socket, $msg, 0, $peerpaddr) )
	    || die "!  Send host: $!";
    };
    logging $@ if $@;
}

sub sendAck {
    my ($socket,$token_id,$token_passing_id,$lost,$peerpaddr)=@_;
    my $flag=1;
    eval {
	logging " > Sending ACKNOWLEDGE packet (%s, %s)\n"
	    ,$token_id, $token_passing_id;
	( $debug == -1 ) && inject_problem();
	my $msg=build_packet("Cnh8h8",$flag,0,$token_id,$token_passing_id);
	( $lost > 0 ) && ( $msg = add_certificate($msg) );	    
	( $debug > 2 ) && dump_packet("ACKNOWLEDGE",$msg);
# Send ACKNOWLEDGE packet
	defined( send($socket, $msg, 0, $peerpaddr) )
	    || die "!  Send host: $!";
    };
    logging $@ if $@;
}

sub sendCommit {
    my ($socket,$token_id,$token_passing_id,$lost,$peerpaddr)=@_;
    my $flag=2;
    eval {
	logging " > Sending COMMIT packet (%s, %s)\n"
	    ,$token_id, $token_passing_id;
	( $debug == -1 ) && inject_problem();
	my $msg=build_packet("Cnh8h8",$flag,0,$token_id,$token_passing_id);
	( $debug > 2 ) && dump_packet("COMMIT",$msg);
# Send ACKNOWLEDGE packet
	defined( send($socket, $msg, 0, $peerpaddr) )
	    || die "!  Send host: $!";
    };
    logging $@ if $@;
}

sub sendEarlyStop {
    my ($socket,$token_id,$token_passing_id,$lost,$peerpaddr)=@_;
    my $flag=3;
    eval {
	logging " > Sending EARLYSTOP packet (%s, %s)\n"
	    ,$token_id, $token_passing_id;
	( $debug == -1 ) && inject_problem();
	my $msg=build_packet("Cnh8h8",$flag,0,$token_id,$token_passing_id);
	( $debug > 2 ) && dump_packet("EARLYSTOP",$msg);
	defined( send($socket, $msg, 0, $peerpaddr) )
	    || die "!  Send host: $!";
    };
    logging $@ if $@;
}

sub token_wait {
    my ( $port,$wait_timeout,$commit_timeout,$retry_timeout ) = @_;

# Socket on wellknown port preparation
    my $socket = mkSocket($port,INADDR_ANY);
    my $timeout=$wait_timeout;
    my $state="waiting";
    my $peerpaddr;
    my $token_id;
    my $token_passing_id;
    my $t0 = [gettimeofday];
    my $t1;

    while ( 1 ) {
# EVENT aquisition
	my $msg;
	my $hispaddr=undef;
	logging "== waiting for incoming packet (state=$state, timeout=$timeout)\n";
	eval {
	    local $SIG{ALRM} = sub { die "Timeout\n" };
	    local $SIG{INT} = sub { die "Terminate\n" };
	    local $SIG{TERM} = sub { die "Terminate\n" };
		local $SIG{__DIE__} = sub {};
	    alarm $timeout;
	    $hispaddr=recv($socket, $msg, 1000, 0);
	    alarm 0;
	};
    if ( $@ =~ /Terminate/ ) {
		logging "! Wait interrupted\n";
		die;
	}
		    
	( $debug > 2 ) && dump_packet("INCOMING",$msg);
	if ( $state eq "waiting" ) {
# Timeout
	    if ( ! defined $hispaddr ) {
		logging "!  Token wait timeout\n";
		close $socket;
		return undef;
	    }
	    my ( $flags, $timing );
	    $peerpaddr=$hispaddr;
	    ( $flags, $timing, $token_id, $token_passing_id )=
		parse_packet ("Cnh8h8",$msg,$hispaddr);
	    if ( ! defined $flags ) {
# An illegal packet
		print "!  Dropping packet\n";
		$timeout=$wait_timeout - tv_interval($t0);
		$state="waiting";
		next;
	    }
	    if ( $flags == 0 ) {
# A MOVE packet
		logging "<  MOVE packet received (%s, %s)\n"
		    ,$token_id, $token_passing_id;
# New socket preparation
		close $socket;
		my $socket;
		while ( 1 ) {
		    $socket = eval { mkSocket(randport(),INADDR_ANY); };
		    logging $@ if $@;
		    last unless $@;
#		    die $@ unless $@=~"Address already in use";
		    logging " !  drawing another port...\n";
#DA SOSTITUIRE CON UN CONTEGGIO
		    sleep 1;
		    next;
		}
		    
# Sending ACKNOWLEDGE
		sendAck($socket,$token_id,$token_passing_id,0,$peerpaddr);
# Enter acknowledge state
		$t1 = [gettimeofday];		
		$timeout=$retry_timeout;
		$state="acknowledge";
		next;
	    }
	    # Anything else
	    $timeout=$retry_timeout;
	    print "!  Dropping packet\n";
	    next;
	}
	if ( $state eq "acknowledge" ) {
# Timeout
	    if ( ! defined $hispaddr ) {
		if ( tv_interval($t1) > $commit_timeout ) {
		    logging "!  Commit timeout; token passing failure\n";
		    close $socket;
		    $timeout=$wait_timeout - tv_interval($t0);
		    $state="waiting";
		    $peerpaddr=undef;
		    $token_id=undef;
		    $token_passing_id=undef;
		    next;
		}	    
		logging "!  Commit timeout\n";
# Sending ACKNOWLEDGE
		sendAck($socket,$token_id,$token_passing_id,1,$peerpaddr);
		$timeout=$retry_timeout;
		next;
	    }
	    my ( $flags, $timing );
	    ( $flags, $timing, $token_id, $token_passing_id )=
		parse_packet ("Cnh8h8",$msg,$peerpaddr);
# Illegal packet
	    if ( ! defined $flags ) {
		logging "!  Dropping packet\n";
		$timeout=$retry_timeout;
		next;
	    }
# Is a COMMIT packet
	    if ( $flags == 2 ) {
		logging "<  COMMIT packet received (%s, %s)\n"
		    ,$token_id, $token_passing_id;
# Sending EARLYSTOP
		sendEarlyStop ($socket,$token_id,$token_passing_id,0,$peerpaddr);
		close $socket;
		my ( $peer_port, $peer_iaddr) = sockaddr_in($peerpaddr);
		my $peer=inet_ntoa($peer_iaddr);
		return ($token_id,$peer,undef);
	    }
# Anything else
	    print "!  Dropping packet\n";   
	    $timeout=$retry_timeout;
	    next;
	}
    }
}
#    return ( $token_id, $waiting_time, $token_latency );
#}

sub token_pass {
    my ( $token_id, $localhost, $peer, $port, $ack_timeout, $retry_timeout ) = @_;

    logging "== Passing token to $peer\n";
# Create socket
    my $iaddr = gethostbyname($localhost);
    my $socket;
    while ( 1 ) {
	$socket= eval { mkSocket(randport(),INADDR_ANY); };
	logging $@ if $@;
	last unless $@;
	die $@ unless $@=~"Address already in use";
	logging " !  drawing another port...\n";
#DA SOSTITUIRE CON UN CONTEGGIO
	sleep 1;
	next;
    }
# Extract peer address
    my $hisiaddr = inet_aton($peer)
	|| (logging "!   Unknown host $peer\n") && (return undef);
    my $peerpaddr = sockaddr_in($port, $hisiaddr);
#Initialize state
    my $timeout=$retry_timeout;
    my $state="moving";
    my $hispaddr;
    my $t0 = [gettimeofday];
    my $t1;
    my $token_passing_id=randid();

    sendMove ($socket,$token_id,$token_passing_id,0,$peerpaddr);
    
    while ( 1 ) {
# EVENT aquisition (token pass)
	my $msg;
	$hispaddr=undef;
	logging "== waiting for incoming packet (state=$state, timeout=$timeout)\n";
	eval {
	    local $SIG{ALRM} = sub { die "Timeout\n" };
	    alarm $timeout;
	    $hispaddr=recv($socket, $msg, 1000, 0);
	    alarm 0;
	};    
	( $debug > 2 ) && dump_packet("INCOMING",$msg);
	if ( $state eq "moving" ) {
	    if ( ! defined $hispaddr ) {
# Timeout
		if ( tv_interval($t0) > $ack_timeout ) {
		    logging "!  Acknowledge timeout; token passing failure\n";
		    close $socket;
		    return undef;
		}	    
		logging "!  Acknowledge wait timeout\n";
		sendMove ($socket,$token_id,$token_passing_id,1,$peerpaddr);
		$timeout=$retry_timeout;
		next;
	    }
	    my ( $flags, $timing, $this_token_id, $this_token_passing_id );
	    ( $flags, $timing, $this_token_id, $this_token_passing_id )=
		parse_packet ("Cnh8h8",$msg,$hispaddr);
	    if ( ! defined $flags ) {
# Illegal packet
		logging "!  Dropping packet\n";
#		sendMove ($socket,$token_id,$token_passing_id,$lost,$peerpaddr);
		$timeout=$retry_timeout;
		next;
	    }
	    if ( $flags == 1 ) {
		logging "<  ACKNOWLEDGE packet received (%s, %s)\n"
		    ,$token_id, $token_passing_id;
		$peerpaddr=$hispaddr;
# Send COMMIT
		sendCommit ($socket,$token_id,$token_passing_id,0,$peerpaddr);
# Enter committing state
		$t1 = [gettimeofday];		
		$timeout=$retry_timeout;
		$state="committing";
		next;
	    }
# Anything else
	    $timeout=$ack_timeout - tv_interval($t0);
	    logging "!  Dropping packet\n";   
	    next;
	}
	if ( $state eq "committing" ) {
	    if ( ! defined $hispaddr ) {
#cambiare introducendo es_timeout
		if ( tv_interval($t1) > $ack_timeout ) {
		    logging "!  EarlyStop timeout; token passed anyway\n";
		    close $socket;
		    return 1;
		}	    
		logging "!  EarlyStop wait timeout\n";
		sendCommit ($socket,$token_id,$token_passing_id,1,$peerpaddr);
		next;
	    }
	    my ( $flags, $timing, $token_passing_id );
	    ( $flags, $timing, $token_id, $token_passing_id )=
		parse_packet ("Cnh8h8",$msg,$hispaddr);
# Illegal packet
	    if ( ! defined $flags ) {
		logging "!  Dropping packet\n";
# Resend COMMIT
#		sendCommit ($socket,$token_id,$token_passing_id,1,$peerpaddr);
		next;
	    }
	    if ( $flags == 1 ) {
		logging "<  ACKNOWLEDGE packet received (%s, %s)\n"
		    ,$token_id, $token_passing_id;
# Send COMMIT
		sendCommit ($socket,$token_id,$token_passing_id,0,$peerpaddr);
# Enter committing state
		$timeout=$retry_timeout;
		$state="committing";
		next;
	    }
	    if ( $flags == 3 ) {
		logging "<  EARLYSTOP packet received (%s, %s)\n"
		    ,$token_id, $token_passing_id;
		close $socket;
		return 1;
	    }
	    next;
	}
    }
    return;
}


sub unique {    
    my ( $stack, $id ) = @_;

    my $insert=1;
    my $suspect_remove=0;
    my $remove=0;

    for ( my $i=0; $i<=$#{@$stack}; $i++ ) {
	logging "   : Compare $id e $stack->[$i]: ";
	if ( $i == 0 ) {
	    if ( $stack->[$i] eq $id ) {
		logging "update\n";
		# Aggiornare il tempo di passaggio
		$insert=0;
		last;
	    }
	}
	if ( $id lt $stack->[$i] ) {
	    logging "suspect\n";
	    $suspect_remove=1;
	    next;
	}
	if ( $id eq $stack->[$i] && $suspect_remove ) {
	    logging "remove $id\n";
	    $remove=1;
	    last;
	}
	print "\n";    
    }
    if ( $remove == 1 ) {
#	logging "   : Remove";
	my @newstack=();
	for my $a ( @$stack ) {
	    ( $a eq $id ) && next;
	    ( ( defined $newstack[0] ) && ( $a eq $newstack[0] ) ) && next;
#	    logging "$a";
	    push @newstack,$a;
	}
	$stack=\@newstack;
    } elsif ( $insert == 1 ) {
	logging "   : Insert $id\n";
	unshift @$stack,$id;
	( defined $stack->[$stacksize-1] ) && ( pop @$stack );
    }
    return $stack;
}

#sub select_peer {
#    opendir DIR,$certdir;
#    my @peers;
#    while ( my $file=readdir DIR ) {
#	( $file =~ /(\d{1,3}?(\.\d{1,3}?){3}?)\.pub/ ) || next;
#	( $1 eq "id_dsa" ) && next;
##	print $1."\n";
#	push @peers,$1;
#    }
#    closedir DIR;
#    ( defined $peers[0] ) || return undef;
#    return $peers[int(rand($#peers+1))];
#}
